@@ -1,4 +1,4 @@
-use std::{io::{Seek, SeekFrom, Read}, fs::{File, OpenOptions}, path::PathBuf, collections::HashMap, sync::{Arc, Mutex, MutexGuard}};
+use std::{io::{Seek, SeekFrom, Read, BufReader}, fs::{File, OpenOptions}, path::PathBuf, collections::HashMap, sync::{Arc, Mutex, MutexGuard}};
 use databuffer::DataBuffer;
 use crate::util::decompress_container_data;
 
@@ -24,7 +24,7 @@ type IdxFileOpt<'a> = Option<&'a mut CacheIndex>;
   For tips on implementing a full-blown Definition Provider, see [`util::DefProvider`].
  */
 pub struct Cache {
-    pub data_file: Arc<Mutex<File>>,
+    pub data_file: Arc<Mutex<BufReader<File>>>,
     indices: HashMap<u8, CacheIndex>
 }
 
@@ -51,17 +51,17 @@ impl Cache {
         let data_file = match OpenOptions::new()
         .read(true)
         .open(&path_buff) {
-            Ok(n) => Arc::from(Mutex::from(n)),
+            Ok(n) => Arc::from(Mutex::from(BufReader::new(n))),
             Err(e) => {
                 println!("Failed opening data file: {:?}, Error: {}", &path_buff, e);
                 return None;
             }
         };
 
-        let num_files = info_file.seek(SeekFrom::End(0)).unwrap() / 6;
+        let num_files = info_file.metadata().unwrap().len() / 6;
         let _ = info_file.seek(SeekFrom::Start(0));
 
-        let mut info = CacheIndex::from(255, 500000, info_file, IdxContainerInfo::new());
+        let mut info = CacheIndex::from(255, 500000, BufReader::new(info_file), IdxContainerInfo::new());
         let mut indices = HashMap::<u8, CacheIndex>::new();
 
         for i in 1..num_files {
@@ -70,7 +70,7 @@ impl Cache {
             path_buff.push(format!("main_file_cache.idx{}",&i));
 
             let file = match OpenOptions::new().read(true).open(&path_buff) {
-                Ok(n) => n,
+                Ok(n) => BufReader::new(n),
                 Err(e) => {
                     println!("Error reading idx {}: {}", i, e);
                     continue;
@@ -110,25 +110,43 @@ impl Cache {
 
 pub struct CacheIndex {
     file_id: u8,
-    file: File,
+    file: BufReader<File>,
     max_container_size: u32,
-    container_info: IdxContainerInfo
+    container_info: IdxContainerInfo,
+    last_archive_id: u32
 }
 
 impl CacheIndex {
-    fn from(file_id: u8, max_size: u32, file: File, container_info: IdxContainerInfo) -> Self {
+    fn from(file_id: u8, max_size: u32, file: BufReader<File>, container_info: IdxContainerInfo) -> Self {
         Self {
             file_id,
             max_container_size: max_size,
             file,
-            container_info
+            container_info,
+            last_archive_id: 0
         }
     }
 
-    fn container_data(&mut self, mut data_file: MutexGuard<File>, container_id: u32) -> Option<Vec<u8>> {
+    fn get_container_by_name_hash(&mut self, hash: u32) -> u32 {
+        match self.container_info.containers.iter().filter(|(_,c)| c.name_hash == hash).last() {
+            Some((c,_)) => *c,
+            None => hash
+        }
+    }
+
+    fn container_data(&mut self, mut data_file: MutexGuard<BufReader<File>>, archive_id: u32) -> Option<Vec<u8>> {
         let mut file_buff: [u8; 520] = [0; 520];
         let mut data: [u8;6] = [0; 6];
-        let _ = self.file.seek(SeekFrom::Start(6 * (container_id as u64)));
+
+        if archive_id <= 1 || self.last_archive_id == 0 {
+            let _ = self.file.seek(SeekFrom::Start(6 * archive_id as u64));
+        } else if self.last_archive_id != archive_id - 1 {
+            let seek_offset = 6 * (archive_id as i64 - (self.last_archive_id as i64 + 1));
+            let _ = self.file.seek_relative(seek_offset);
+        }
+
+        self.last_archive_id = archive_id;
+
         let _ = match self.file.read(&mut data) {
             Ok(_) => {}
             Err(e) => {
@@ -136,76 +154,88 @@ impl CacheIndex {
             }
         };
 
-        let container_size = ((data[2] & 0xff) as u32) + ((((0xff & data[0]) as u32) << 16) + (((data[1] as u32) << 8) & 0xff00));
-        let mut sector = (((data[3] & 0xff) as i32) << 16) - (-((0xff & data[4] as i32) << 8) - (data[5] as i32 & 0xff)); 
+        let container_size = (data[2] as u32) + (((data[0] as u32) << 16) + (((data[1] as u32) << 8) & 0xff00));
+        let mut sector = ((data[3] as i32) << 16) - (-((0xff & data[4] as i32) << 8) - (data[5] as i32 & 0xff)); 
 
         if container_size > self.max_container_size {
             println!("Container Size greater than Max Container Size! {} > {}", container_size, self.max_container_size);
             None
+        } else if sector <= 0 {
+            println!("Sector <= 0! {}", sector);
+            None
         } else {
-            if sector <= 0 {
-                println!("Sector <= 0! {}", sector);
-                None
-            } else {
-                let mut container_data = Vec::<u8>::new();
+            let mut container_data = Vec::<u8>::new();
 
-                let mut data_read_count = 0;
-                let mut part = 0;
+            let mut data_read_count = 0;
+            let mut part: u32 = 0;
 
-                while container_size > data_read_count {
-                    if sector == 0 {
-                        println!("Sector == 0!");
-                        return None;
-                    }
+            let initial_dfile_pos = data_file.seek(SeekFrom::Start(520 * (sector as u64))).unwrap() as i64;
 
-                    let _ = data_file.seek(SeekFrom::Start(520 * (sector as u64)));
-
-                    let mut data_to_read = container_size - data_read_count;
-
-                    if data_to_read > 512 {
-                        data_to_read = 512;
-                    }
-
-                    let _ = data_file.read(&mut file_buff);
-
-                    let current_container_id = (0xff & file_buff[1] as u32) + ((0xff & file_buff[0] as u32) << 8);
-                    let current_part = ((0xff & file_buff[2] as u32) << 8) + (0xff & file_buff[3] as u32);
-                    let next_sector = (0xff & file_buff[6] as u32) + ((0xff & file_buff[5] as u32) << 8) + ((0xff & file_buff[4] as u32) << 16);
-                    let current_idx_file_id = 0xff & file_buff[7] as u32;
-
-                    if container_id != (current_container_id as u32) || current_part != part || self.file_id != (current_idx_file_id as u8) {
-                        println!("Multipart failure! {} != {} || {} != {} || {} != {}", container_id, current_container_id, current_part, part, self.file_id, current_idx_file_id);
-                        return None;
-                    }
-
-                    let upper_bound = 8 + data_to_read as usize;
-
-                    container_data.extend_from_slice(&file_buff[8..upper_bound]);
-                    data_read_count += data_to_read;
-
-                    part += 1;
-                    sector = next_sector as i32;
+            while container_size > data_read_count {
+                if sector == 0 {
+                    println!("Sector == 0!");
+                    return None;
                 }
 
-                Some(container_data)
+                let seek_target: i64 = 520 * (sector as i64);
+                let current_pos = initial_dfile_pos + (data_read_count as i64) + (part as i64 * 8);
+
+                if current_pos != seek_target {
+                    let _ = data_file.seek_relative(520);
+                }
+
+                let mut data_to_read = container_size - data_read_count;
+
+                if data_to_read > 512 {
+                    data_to_read = 512;
+                }
+
+                let bytes_read = data_file.read(&mut file_buff).unwrap();
+
+                if data_to_read + 8 > bytes_read as u32 {
+                    let _ = data_file.seek(SeekFrom::Start(520 * (sector as u64)));
+
+                    let _ = data_file.read(&mut file_buff);
+                }
+
+                let current_container_id = (0xff & file_buff[1] as u32) + ((0xff & file_buff[0] as u32) << 8);
+                let current_part = ((0xff & file_buff[2] as u32) << 8) + (0xff & file_buff[3] as u32);
+                let next_sector = (0xff & file_buff[6] as u32) + ((0xff & file_buff[5] as u32) << 8) + ((0xff & file_buff[4] as u32) << 16);
+                let current_idx_file_id = 0xff & file_buff[7] as u32;
+
+                if archive_id != (current_container_id as u32) || current_part != part || self.file_id != (current_idx_file_id as u8) {
+                    println!("Multipart failure! {} != {} || {} != {} || {} != {}", archive_id, current_container_id, current_part, part, self.file_id, current_idx_file_id);
+                    return None;
+                }
+
+                let upper_bound = 8 + data_to_read as usize;
+
+                container_data.extend_from_slice(&file_buff[8..upper_bound]);
+                data_read_count += data_to_read;
+
+                part += 1;
+                sector = next_sector as i32;
             }
+
+            Some(container_data)
         }
     }
 
     pub fn get_total_files(&mut self) -> u32 {
-        self.container_info.container_indices.sort();
+        self.container_info.container_indices.sort_unstable();
 
-        let last_archive_id = self.container_info.container_indices.last().unwrap().clone();
+        let last_archive_id = *self.container_info.container_indices.last().unwrap();
         let last_archive = self.container_info.containers.get(&last_archive_id).unwrap();
 
         let last_archive_file_amount = last_archive.file_indices.len();
         let other_file_amounts = (self.container_info.container_indices.len() - 1) * 256;
         
-        return (last_archive_file_amount + other_file_amounts) as u32;
+        (last_archive_file_amount + other_file_amounts) as u32
     }
 }
 
 #[allow(dead_code)]
+#[derive(Default)]
 pub struct IdxContainerInfo {
     protocol: u8,
     revision: u32,
@@ -217,19 +247,12 @@ pub struct IdxContainerInfo {
 
 impl IdxContainerInfo {
     pub fn new() -> Self {
-        Self {
-            protocol: 0,
-            revision: 0,
-            container_indices: Vec::new(),
-            containers: HashMap::new(),
-            named_files: false,
-            whirlpool: false,
-        }
+        Self::default()
     }
 
     pub fn from(packed_data: Vec<u8>) -> Self {
         let mut data = match decompress_container_data(packed_data) {
-            Some(n) => DataBuffer::from_bytes(&n),
+            Some(n) => DataBuffer::with_vec(n),
             None => {
                 println!("Unable to decompress container data.");
                 return Self::new();
@@ -240,7 +263,7 @@ impl IdxContainerInfo {
         
         if protocol != 5 && protocol != 6 {
             println!("Invalid protocol while parsing container info: {}", protocol);
-            return Self::new();
+            Self::new()
         } else {
             let revision = match protocol {
                 5 => 0,
@@ -258,61 +281,60 @@ impl IdxContainerInfo {
             for i in 0..num_indices {
                 container_indices.push((data.read_u16() as u32) + match i {
                     0 => 0,
-                    _ => container_indices.last().unwrap().clone()
+                    _ => *container_indices.last().unwrap()
                 });
 
-                containers.insert(container_indices.last().unwrap().clone(), IdxContainer::new());
+                containers.insert(*container_indices.last().unwrap(), IdxContainer::new());
             }
 
             if files_named {
-                for i in 0..(num_indices as usize) {
-                    containers.get_mut(&container_indices[i]).unwrap().name_hash = data.read_u32();
+                for c in container_indices.iter().take(num_indices as usize) {
+                    containers.get_mut(c).unwrap().name_hash = data.read_u32();
                 }
             }
 
             let mut file_hashes: HashMap<u32, [u8;64]> = HashMap::new();
 
             if whirlpool {
-                for i in 0..(num_indices as usize) {
+                for c in container_indices.iter().take(num_indices as usize) {
                     let mut buf: [u8; 64] = [0; 64];
                     let _ = data.read(&mut buf);
-                    file_hashes.insert(container_indices[i].clone(), buf);
+                    file_hashes.insert(*c, buf);
                 }
             }
 
-            for i in 0..(num_indices as usize) {
-                let container = containers.get_mut(&container_indices[i]).unwrap();
+            for c in container_indices.iter().take(num_indices as usize) {
+                let container = containers.get_mut(c).unwrap();
                 container.crc = data.read_i32();
             }
 
-            for i in 0..(num_indices as usize) {
-                let container = containers.get_mut(&container_indices[i]).unwrap();
+            for c in container_indices.iter().take(num_indices as usize) {
+                let container = containers.get_mut(c).unwrap();
                 container.version = data.read_i32();
             }
 
             let mut container_index_counts = HashMap::<u32, u16>::new(); 
 
-            for i in 0..(num_indices as usize) {
-                container_index_counts.insert(container_indices[i].clone(), data.read_u16());
+            for c in container_indices.iter().take(num_indices as usize) {
+                container_index_counts.insert(*c, data.read_u16());
             }
 
-            for i in 0..(num_indices as usize) {
-                let container = containers.get_mut(&container_indices[i]).unwrap();
+            for c in container_indices.iter().take(num_indices as usize) {
+                let container = containers.get_mut(c).unwrap();
                 
-                for f in 0..(container_index_counts.get(&container_indices[i]).unwrap().clone() as usize){
+                for f in 0..(*container_index_counts.get(c).unwrap() as usize){
                     container.file_indices.push((data.read_u16() as u32) + match f {
                         0 => 0,
                         _ => container.file_indices[f - 1]
                     });
 
-                    container.file_containers.insert(container.file_indices[f].clone(), IdxFileContainer::new());
+                    container.file_containers.insert(container.file_indices[f], IdxFileContainer::new());
                 }
             }
 
             if whirlpool {
-                for container_index in 0..container_indices.len() {
+                for (container_index, container_id) in container_indices.iter().enumerate() {
                     for file_index in 0..containers.get(&(container_index as u32)).unwrap().file_containers.len() {
-                        let container_id = container_indices[container_index];
                         let file_id = containers.get(&container_id).unwrap().file_indices[file_index];
                         
                         containers.get_mut(&container_id).unwrap()
@@ -323,8 +345,8 @@ impl IdxContainerInfo {
             }
 
             if files_named {
-                for i in 0..(num_indices as usize) {
-                    let container = containers.get_mut(&container_indices[i]).unwrap();
+                for c in container_indices.iter().take(num_indices as usize) {
+                    let container = containers.get_mut(c).unwrap();
 
                     for f in 0..(container.file_indices.len()) {
                         let file = container.file_containers.get_mut(&container.file_indices[f]).unwrap();
@@ -346,6 +368,7 @@ impl IdxContainerInfo {
     }
 }
 
+#[derive(Default)]
 pub struct IdxContainer {
     version: i32,
     name_hash: u32,
@@ -356,17 +379,12 @@ pub struct IdxContainer {
 
 impl IdxContainer {
     pub fn new() -> Self {
-        Self {
-            version: -1,
-            name_hash: 0,
-            crc: -1,
-            file_indices: Vec::new(),
-            file_containers: HashMap::new()
-        }
+        Self::default()
     }
 }
 
 #[allow(dead_code)]
+#[derive(Default)]
 pub struct IdxFileContainer {
     version: u8,
     name_hash: u32,
@@ -376,11 +394,6 @@ pub struct IdxFileContainer {
 
 impl IdxFileContainer {
     pub fn new() -> Self {
-        Self {
-            version: 0,
-            name_hash: 0,
-            crc: -1,
-            data: Vec::new()
-        }
+        Self::default()
     }
 }
\ No newline at end of file
